# We use the construction of polyharmonic weak Maass forms from spectral
# families. Notation following the paper: k the weight of the family, m the
# symmetric power, d the desired depth.

################################################################################
################################################################################
# auxiliary functions
################################################################################
################################################################################

function pochhammer(a::A, n::Int) :: A where {A <: RingElem}
  @assert n >= 0
  if n == 0
    return one(parent(a))
  end
  return prod(a+i for i in 0:n-1)
end

# The alternating trace
function tdtr(v::Vector{A}) :: A where {A <: RingElem}
  @assert !isempty(v)
  sum((-1)^(i-1)*a for (i,a) in enumerate(v))
end

################################################################################
################################################################################
# construction of Δ and its action on spectral families
################################################################################
################################################################################

# The triple A, B C that corresponds via the transformaton A + T B + T^2 C to
# Delta acting on f_L.
function delta_on_Lfamily(k::Int, m::Int) :: DeltaSpectralFamily
  @assert m >= 0

  A = zeros(QQ, m+1,m+1)
  B = zeros(QQ, m+1,m+1)
  C = zeros(QQ, m+1,m+1)
  for r in 0:m
    A[r+1,r+1] = (m-r)*(m-2*r-k)
    B[r+1,r+1] = (1-k)
    C[r+1,r+1] = -1
    if r > 0
      A[r+1-1,r+1] = -1
    end
    if r < m
      A[r+1+1,r+1] = (r+1)*(m-r)*(m-r-1)*(m-r-k)
      B[r+1+1,r+1] = (r+1)*(m-r)*(1-k)
      C[r+1+1,r+1] = -(r+1)*(m-r)
    end
  end

  # Construct the pseudo inverse by finding the pivot elements of A. Since the
  # image has vanishing last coordinate, we can use row indices 1:m.
  Am = matrix(QQ,A)
  Arr = rref(Am)[2]
  Apv = Int[first([j for j in 1:m+1 if !iszero(Arr[i,j])]) for i in 1:m]
  Apvinv = inv(Am[1:m,Apv])
  Apsdinv = zero_matrix(QQ,m+1,m+1)
  for i in 1:m, (j,jp) in zip(Apv,1:m)
    Apsdinv[i,j] = Apvinv[i,jp]
  end

  return DeltaSpectralFamily(:L,k,m,A,B,C,Matrix(Apsdinv))
end

# The triple A, B C that corresponds via the transformaton A + T B + T^2 C to
# Delta acting on f_L.
function delta_on_Rfamily(k::Int, m::Int) :: DeltaSpectralFamily
  @assert m >= 0

  A = zeros(QQ, m+1,m+1)
  B = zeros(QQ, m+1,m+1)
  C = zeros(QQ, m+1,m+1)
  for r in 0:m
    A[r+1,r+1] = -(r*(m-2*r-k)+m)
    B[r+1,r+1] = (1-k)
    C[r+1,r+1] = -1
    if r > 0
      A[r+1-1,r+1] = r*(r-1+k)
      B[r+1-1,r+1] = (1-k)
      C[r+1-1,r+1] = -1
    end
    if r < m
      A[r+1+1,r+1] = -(r+1)*(m-r)
    end
  end

  # Construct the pseudo inverse by finding the pivot elements of A. Since the
  # image has vanishing alternating trace, we can use row indices 1:m.
  Am = matrix(QQ,A)
  Arr = rref(Am)[2]
  Apv = Int[first([j for j in 1:m+1 if !iszero(Arr[i,j])]) for i in 1:m]
  Apvinv = inv(Am[1:m,Apv])
  Apsdinv = zero_matrix(QQ,m+1,m+1)
  for i in 1:m, (j,jp) in zip(Apv,1:m)
    Apsdinv[i,j] = Apvinv[i,jp]
  end

  return DeltaSpectralFamily(:R,k,m,A,B,C,Matrix(Apsdinv))
end

# The action of Delta on v_0 + T v_1 ... encoded as a matrix, whose column
# vectors are v_0, v_1, ....
function *(
    delta::DeltaSpectralFamily,
    vs::Matrix{A}
   ) :: Matrix{A} where {
    A <: RingElem
   }
  m = sympower(delta)
  @assert size(vs,1) == m+1

  Avs = delta.A*vs
  if size(vs,1) >= 1
    Bvs = hcat(zeros(QQ,m+1,1), delta.B*vs[:,1:end-1])
  else
    Bvs = zeros(QQ,m+1,0)
  end
  if size(vs,1) >= 2
    Cvs = hcat(zeros(QQ,m+1,2), delta.C*vs[:,1:end-2])
  else
    Cvs = zeros(QQ,m+1,size(Bvs,2))
  end

  return Avs .+ Bvs .+ Cvs
end

function apply_delta(f::LRConstrLRExpansion) :: LRConstrLRExpansion
  return lrconstuction_from_coefficients_pullback(delta(f),
           delta(f) * coefficients_pullback(f))
end

delta(f::LRConstrLRExpansion) = f.delta

function preimage_A(v::Vector{fmpq}, delta::DeltaSpectralFamily) :: Vector{fmpq}
  m = delta.m
  @assert length(v) == m+1
  # We do not check that v lies in the image of A.
  return delta.Apsdinv*v
end

islconstr(delta::DeltaSpectralFamily) = delta.LR === :L
isrconstr(delta::DeltaSpectralFamily) = delta.LR === :R

base_weight(delta::DeltaSpectralFamily) = delta.k

sympower(delta::DeltaSpectralFamily) = delta.m

function weight(delta::DeltaSpectralFamily)
  if islconstr(delta)
    return base_weight(delta) - sympower(delta)
  else
    return base_weight(delta) + sympower(delta)
  end
end

################################################################################
################################################################################
# L and R construction of polyharmonic Maaß forms
################################################################################
################################################################################

sympower(f::LRConstrLRExpansion) = size(coefficients(f),1)-1

depth(f::LRConstrLRExpansion) = size(coefficients(f),2)-1

base_weight(f::LRConstrLRExpansion) = base_weight(delta(f))

weight(f::LRConstrLRExpansion) = weight(delta(f))

islconstr(f::LRConstrLRExpansion) = islconstr(delta(f))
isrconstr(f::LRConstrLRExpansion) = isrconstr(delta(f))

function show(io::IO, mime::MIME"text/plain", f::LRConstrLRExpansion)
  d = depth(f)
  m = sympower(f)
  fc = coefficients(f)
  if islconstr(f)
    show(io, mime, [iszero(fc[r+1,t+1]) ? "0" : "$(fc[r+1,t+1]) 𝔢_{$r,$(m-r)} L^{$(m-r)} f^{($t)}_{0,0}"
               for r in 0:m, t in 0:d])
  elseif isrconstr(f)
    show(io, mime, [iszero(fc[r+1,t+1]) ? "0" : "$(fc[r+1,t+1]) 𝔢_{$r,$(m-r)} R^{$r} f^{($t)}_{0,0}"
               for r in 0:m, t in 0:d])
  else
    error()
  end
end

function coefficients_pullback(f::LRConstrLRExpansion) :: Matrix{fmpq}
  m = sympower(f)
  d = depth(f)
  fc = coefficients(f)
  return [fc[r+1,d-t+1] * factorial(ZZ(d-t)) for r in 0:m, t in 0:d]
end

# Given an expansion v_0 + T v_1 + ⋯ that corresponds to an L or R
# construction, renormalize via the pullback along ϕ in the paper.
function lrconstuction_from_coefficients_pullback(
    delta::DeltaSpectralFamily,
    coeffs::Matrix{fmpq}
   ) :: LRConstrLRExpansion
  m = size(coeffs,1)-1
  d = size(coeffs,2)-1
  # We want to use the coefficients in increasing order of t in f^{(t)}, while
  # the coefficients that we receive are in increasing order of t in T^t.
  return LRConstrLRExpansion(delta,
           [coeffs[r+1,d-t+1] // factorial(ZZ(t)) for r in 0:m, t in 0:d])
end

# The L-construction for the m-th symmetric power in depth d of a weight k
# spectral family.
function lconstruction(k::Int, m::Int, d::Int) :: LRConstrLRExpansion
  @assert m >= 0
  @assert d >= 0
  delta = delta_on_Lfamily(k,m)

  w0 = zeros(QQ,m+1,1)
  if m >= k
    for r in 0:min(m,m-k)
      w0[r+1] = inv(QQ(factorial(ZZ(m-r))*factorial(ZZ(m-r-k))))
    end
  else # m < k
    for r in 0:m
      w0[r+1,1] = inv(QQ(factorial(ZZ(m-r))*pochhammer(ZZ(1-k),m-r)))
    end
  end

  if d == 0
    return LRConstrLRExpansion(delta, w0)
  end
  @assert k <= 0 || k > m+1
  wt = w0
  # We extend wt at the end, apply Delta and then determine the preimage under
  # A of the last component after reducing with the already available vectors.
  # c0 is a normalization constant that we need in the reduction process.
  c0 = inv(w0[m+1,1])

  for t in 1:d
    # This is wt without the last coefficient T^t v.
    wt = hcat(wt, zeros(QQ,m+1,1))
    delta_wt_TtAv = (delta*wt)[:,2:end]
    # We reduce all components of delta_wt_TtAv by the previous wt, which are
    # the intitial parts of wt. We use the depth filtration and the fact that
    # the top nonzero component must be a multiple of w0, whose multiples are
    # determined by their last coordinate.
    for tt in 1:t
      ctt = c0 * delta_wt_TtAv[m+1,tt]
      delta_wt_TtAv[:,tt:end] .-= ctt .* wt[:,1:end-tt]
    end
    # The resulting element is in the image of A
    @assert iszero(delta_wt_TtAv[:,1:end-1])
    @assert iszero(delta_wt_TtAv[m+1,end])
    wt[:,end] .= -preimage_A(delta_wt_TtAv[:,end], delta)
  end

  return lrconstuction_from_coefficients_pullback(delta,wt)
end

# The R-construction for the m-th symmetric power in depth d of a weight k
# spectral family.
function rconstruction(k::Int, m::Int, d::Int) :: LRConstrLRExpansion
  @assert m >= 0
  @assert d >= 0
  delta = delta_on_Rfamily(k,m)

  w0 = zeros(QQ,m+1,1)
  if m > -k
    for r in max(0,1-k):m
      w0[r+1,1] = inv(QQ(factorial(ZZ(m-r))*factorial(ZZ(r+k-1))))
    end
  else # m \le -k
    for r in 0:m
      w0[r+1,1] = inv(QQ(factorial(ZZ(m-r))*pochhammer(ZZ(k),r)))
    end
  end

  if d == 0
    return LRConstrLRExpansion(delta, w0)
  end
  @assert k != 1
  wt = w0
  # We extend wt at the end, apply delta and then determine the preimage under
  # A of the last component after reducing with the already available vectors.
  # c0 is a normalization constant that we need in the reduction process.
  c0 = inv(tdtr(w0[:,1]))

  for t in 1:d
    # This is wt without the last coefficient T^t v.
    wt = hcat(wt, zeros(QQ,m+1,1))
    delta_wt_TtAv = (delta*wt)[:,2:end]
    # We reduce all components of delta_wt_TtAv by the previous wt, which are
    # the intitial parts of wt. We use the depth filtration and the fact that
    # the top nonzero component must be a multiple of w0, whose multiples are
    # determined by their alternating trace.
    for tt in 1:t
      ctt = c0 * tdtr(delta_wt_TtAv[:,tt])
      delta_wt_TtAv[:,tt:end] .-= ctt .* wt[:,1:end-tt]
    end
    # The resulting element is in the image of A
    @assert iszero(delta_wt_TtAv[:,1:end-1])
    @assert iszero(tdtr(delta_wt_TtAv[:,end]))
    wt[:,end] .= -preimage_A(delta_wt_TtAv[:,end], delta)
  end

  return lrconstuction_from_coefficients_pullback(delta,wt)
end

coefficients(f::LRConstrLRExpansion) = f.coeffs

iszero(f::LRConstrLRExpansion) = iszero(coefficients(f))
