# Expansion of the L and R construction as specializations of spectral
# families.

base_ring(f::LRConstrSpectralPointExpansion) = f.base_ring

poly_ring(f::LRConstrSpectralPointExpansion) = f.poly_ring

base_weight(f::LRConstrSpectralPointExpansion) = f.k

weight_shift(f::LRConstrSpectralPointExpansion) = f.weight_shift

spectral_shift(f::LRConstrSpectralPointExpansion) = f.spectral_shift

depth(f::LRConstrSpectralPointExpansion) = size(f.coeffs,2)-1

sympower(f::LRConstrSpectralPointExpansion) = size(f.coeffs,1)-1

lowering_poly(f::LRConstrSpectralPointExpansion) = f.lpoly

raising_poly(f::LRConstrSpectralPointExpansion) = f.rpoly

mirror_scale(f::LRConstrSpectralPointExpansion) = f.mirror_scale

mirror_sign(f::LRConstrSpectralPointExpansion) = f.mirror_sign

coefficients(f::LRConstrSpectralPointExpansion) = f.coeffs

islconstr(f::LRConstrSpectralPointExpansion) = f.LR === :L

isrconstr(f::LRConstrSpectralPointExpansion) = f.LR === :R

iszero(f::LRConstrSpectralPointExpansion) = iszero(coefficients(f))

function ==(
    f::LRConstrSpectralPointExpansion,
    g::LRConstrSpectralPointExpansion
   )
  base_ring(f)      == base_ring(g)      || return false
  poly_ring(f)      == poly_ring(g)      || return false
  base_weight(f)    == base_weight(g)    || return false
  weight_shift(f)   == weight_shift(g)   || return false
  spectral_shift(f) == spectral_shift(g) || return false
  depth(f)          == depth(g)          || return false
  sympower(f)       == sympower(g)       || return false
  return coefficients(f) == coefficients(g)
end

function show(io::IO, mime::MIME"text/plain", f::LRConstrSpectralPointExpansion)
  k = base_weight(f)
  l = weight_shift(f)
  s0 = spectral_shift(f)
  d = depth(f)
  m = sympower(f)
  fc = coefficients(f)
  if islconstr(f)
    show(io, mime,
         [iszero(fc[r+1,t+1]) ? "0" :
            "$(fc[r+1,t+1]) 𝔢_{$r,$(m-r)} f^{($t)}_{$(k+2*l-2*m+2*r),$(-l+m-r+s0)}"
          for r in 0:m, t in 0:d])
  elseif isrconstr(f)
    show(io, mime,
         [iszero(fc[r+1,t+1]) ? "0" :
            "$(fc[r+1,t+1]) 𝔢_{$r,$(m-r)} f^{($t)}_{$(k+2*l+2*r),$(-l-r+s0)}"
          for r in 0:m, t in 0:d])
  else
    error()
  end
end

function -(f::LRConstrSpectralPointExpansion)
  return LRConstrSpectralPointExpansion(f.base_ring, f.poly_ring, 
           f.LR, f.k, f.weight_shift, f.spectral_shift,
           f.lpoly, f.rpoly, f.mirror_scale, f.mirror_sign,
           -f.coeffs)
end

################################################################################
################################################################################
# construction
################################################################################
################################################################################

# Convert a L or R construction to its representation via specializations at
# various spectral points of the Eisenstein series.
function eisenstein_spectral_point_expansion(f::LRConstrLRExpansion)
  base_ring = QQ
  poly_ring,s = PolynomialRing(QQ, "s")
  lpoly(k,r) = r == 0 ? one(poly_ring) : prod(s+i for i in 0:r-1)
  rpoly(k,r) = r == 0 ? one(poly_ring) : prod(s+k+i for i in 0:r-1)

  return _spectral_point_expansion(f, base_ring, poly_ring,
             lpoly, rpoly, one(QQ), 1)
end

# Convert a L or R construction to its representation via specializations at
# various spectral points of the Poincaré series.
function poincare_spectral_point_expansion(f::LRConstrLRExpansion, n::Int)
  @assert n != 0
  n = abs(n)

  base_ring,pi = LaurentPolynomialRing(QQ, "pi")
  poly_ring,s = PolynomialRing(base_ring, "s")
  lpoly(k,r) = r == 0 ? one(poly_ring) : inv(4*pi*n)^r * prod(1-k-s+i for i in 0:r-1)
  rpoly(k,r) = r == 0 ? one(poly_ring) : (4*pi*n)^r * prod(1-s+i for i in 0:r-1)
  mirror_scale = 4*pi*n
  mirror_sign = iseven(weight(f)) ? -1 : 1

  return _spectral_point_expansion(f, base_ring, poly_ring,
             lpoly, rpoly, mirror_scale, mirror_sign)
end

function _spectral_point_expansion(
    f::LRConstrLRExpansion,
    base_ring::Ring,
    poly_ring::PolyRing,
    lpoly::Function,
    rpoly::Function,
    mirror_scale::RingElem,
    mirror_sign::Int
   ) :: LRConstrSpectralPointExpansion
  k = base_weight(f)
  m = sympower(f)
  d = depth(f)
  coeffs = zeros(base_ring, m+1,d+1)
  fcoeffs = coefficients(f)
  if islconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} L^{m-r} f^{(t)}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   ∂_{s=0}^t (lpoly(k,m-r)(s) * f_{s+m-r})
      # = \sum_{tt=0}^t binomial(t,tt)
      #     ( ∂_{s=0}^{tt} lpoly(k,m-r) )
      #   * ( ∂_{s=0}^{t-tt} f_{s+m-r}) )
      p = lpoly(k,m-r)
      for tt in 0:t
        coeffs[r+1,t-tt+1] += binomial(t,tt) * p(0) * fc
        p = derivative(p)
      end
    end
  elseif isrconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} R^r f^{(t)}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   ∂_{s=0}^t (rpoly(k,r)(s) * f_{s-r})
      # = \sum_{tt=0}^t binomial(t,tt)
      #     ( ∂_{s=0}^{tt} rpoly(k,r) )
      #   * ( ∂_{s=0}^{t-tt} f_{s-r}) )
      p = rpoly(k,r)
      for tt in 0:t
        coeffs[r+1,t-tt+1] += binomial(t,tt) * p(0) * fc
        p = derivative(p)
      end
    end
  else
    error()
  end

  return LRConstrSpectralPointExpansion(base_ring, poly_ring, 
           PolyHarmonic.delta(f).LR, k, 0, 0,
           lpoly, rpoly, mirror_scale, mirror_sign,
           coeffs)
end

################################################################################
################################################################################
# differential operators
################################################################################
################################################################################

function laplace(f::LRConstrSpectralPointExpansion, n::Int = 1)
  n == 0 && return f
  @assert n > 0

  for _ in 1:n
    f = -raising(lowering(f))
  end
  return f
end

function lowering(f::LRConstrSpectralPointExpansion, n::Int = 1)
  n == 0 && return f
  @assert n > 0

  k = base_weight(f)
  l = weight_shift(f)
  d = depth(f)
  m = sympower(f)
  s0 = spectral_shift(f)
  lpoly = lowering_poly(f)
  coeffs = zeros(base_ring(f), m+1,d+1)
  fcoeffs = coefficients(f)

  if islconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{m-r-l+s0}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   L^n (𝔢_{r,m-r} f)
      # = \sum_{rr=r}^m binomial(n,rr-r)
      #                 rr! / r! (m-r)! / (m-rr)!
      #                 𝔢_{rr,m-rr} L^{n-(rr-r)} f
      # and
      #   ∂_{s=m-r-l+s0}^t lpoly(k+2l-m-(m-2r),n-(rr-r))(s) * f_{s+(n-(rr-r))}
      # = \sum_{tt=0}^t binomial(t,tt)
      #     ( ∂_{s=m-r-l+s0}^{tt}   lpoly(k+2l-m-(m-2r),n-(rr-r)) )
      #   * ( ∂_{s=m-r-l+s0}^{t-tt} f_{s+(n-(rr-r))}) )
      for rr in r:min(m,r+n)
        p = binomial(n,rr-r) *
            factorial(ZZ(rr))÷factorial(ZZ(r)) *
            factorial(ZZ(m-r))÷factorial(ZZ(m-rr)) *
            lpoly(k+2*l-m-(m-2*r),n-(rr-r))
        for tt in 0:t
          coeffs[rr+1,t-tt+1] += binomial(t,tt) * p(m-r-l+s0) * fc
          p = derivative(p)
        end
      end
    end
  elseif isrconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{-r-l+s0}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   L^n (𝔢_{r,m-r} f)
      # = \sum_{rr=r}^m binom(n,rr-r)
      #                 rr! / r! (m-r)! / (m-rr)!
      #                 𝔢_{rr,m-rr} L^{n-(rr-r)} f
      # and
      #   ∂_{s=-r-l+s0}^t lpoly(k+2l+m-(m-2r),n-(rr-r))(s) * f_{s+(n-(rr-r))}
      # = \sum_{tt=0}^t binom(t,tt)
      #     ( ∂_{s=-r-l+s0}^{tt}   lpoly(k+2l+m-(m-2r),n-(rr-r)) )
      #   * ( ∂_{s=-r-l+s0}^{t-tt} f_{s-(n-rr)}) )
      for rr in r:min(m,r+n)
        p = binomial(n,rr-r) *
            factorial(ZZ(rr))÷factorial(ZZ(r)) *
            factorial(ZZ(m-r))÷factorial(ZZ(m-rr)) *
            lpoly(k+2*l+m-(m-2*r),n-(rr-r))
        for tt in 0:t
          coeffs[rr+1,t-tt+1] += binomial(t,tt) * p(-r-l+s0) * fc
          p = derivative(p)
        end
      end
    end
  else
    error()
  end

  return LRConstrSpectralPointExpansion(f.base_ring, f.poly_ring, 
           f.LR, k, l-n, s0,
           f.lpoly, f.rpoly, f.mirror_scale, f.mirror_sign,
           coeffs)
end

function raising(f::LRConstrSpectralPointExpansion, n::Int = 1)
  n == 0 && return f
  @assert n > 0

  k = base_weight(f)
  l = weight_shift(f)
  d = depth(f)
  m = sympower(f)
  s0 = spectral_shift(f)
  rpoly = raising_poly(f)
  coeffs = zeros(base_ring(f), m+1,d+1)
  fcoeffs = coefficients(f)

  if islconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{m-r-l+s0}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   R^n (𝔢_{r,m-r} f)
      # = \sum_{rr=0}^r binomial(n,r-rr) 𝔢_{rr,m-rr} R^{n-(r-rr)} f
      # and
      #   ∂_{s=m-r-l+s0}^t rpoly(k+2l-m-(m-2r),n-(r-rr))(s) * f_{s-(n-(r-rr))}
      # = \sum_{tt=0}^t binomial(t,tt)
      #     ( ∂_{s=m-r-l+s0}^{tt}   rpoly(k+2l-m-(m-2r),n-(r-rr)) )
      #   * ( ∂_{s=m-r-l+s0}^{t-tt} f_{s-(n-(r-rr))}) )
      for rr in max(0,r-n):r
        p = binomial(n,r-rr) * rpoly(k+2*l-m-(m-2*r),n-(r-rr))
        for tt in 0:t
          coeffs[rr+1,t-tt+1] += binomial(t,tt) * p(m-r-l+s0) * fc
          p = derivative(p)
        end
      end
    end
  elseif isrconstr(f)
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{-r-l+s0}
      fc = fcoeffs[r+1,t+1]
      # We have
      #   R^n (𝔢_{r,m-r} f)
      # = \sum_{rr=0}^r binom(n,r-rr) 𝔢_{rr,m-rr} R^{n-(r-rr)} f
      # and
      #   ∂_{s=-r-l+s0}^t rpoly(k+2l+m-(m-2r),n-(r-rr))(s) * f_{s-(n-(r-rr))}
      # = \sum_{tt=0}^t binom(t,tt)
      #     ( ∂_{s=-r-l+s0}^{tt}   rpoly(k+2l+m-(m-2r),n-(r-rr)) )
      #   * ( ∂_{s=-r-l+s0}^{t-tt} f_{s-(n-rr)}) )
      for rr in max(0,r-n):r
        p = binomial(n,r-rr) * rpoly(k+2*l+m-(m-2*r),n-(r-rr))
        for tt in 0:t
          coeffs[rr+1,t-tt+1] += binomial(t,tt) * p(-r-l+s0) * fc
          p = derivative(p)
        end
      end
    end
  else
    error()
  end

  return LRConstrSpectralPointExpansion(f.base_ring, f.poly_ring, 
           f.LR, k, l+n, s0,
           f.lpoly, f.rpoly, f.mirror_scale, f.mirror_sign,
           coeffs)
end

function mirror(f::LRConstrSpectralPointExpansion)
  k = base_weight(f)
  l = weight_shift(f)
  d = depth(f)
  m = sympower(f)
  s0 = spectral_shift(f)
  msc = mirror_scale(f)
  msgn = mirror_sign(f)
  coeffs = zeros(base_ring(f), m+1,d+1)
  fcoeffs = coefficients(f)

  if islconstr(f)
    LR = :R
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{-r-l+s0}
      # We have
      #   y^{m-2r} \ov{𝔢_{r,m-r}}
      # = (-1)^m (m-r)! / r! 𝔢_{m-r,r}
      coeffs[m-r+1,t+1] = ( msgn * msc^-(k+2*l-m-(m-2*r))
                          * (-1)^m * factorial(ZZ(m-r)) // factorial(ZZ(r))
                          * fcoeffs[r+1,t+1] )
    end
  elseif isrconstr(f)
    LR = :L
    for r in 0:m, t in 0:d
      # The contribution of 𝔢_{r,m-r} f^{(t)}_{-r-l+s0}
      # We have
      #   y^{m-2r} \ov{𝔢_{r,m-r}}
      # = (-1)^m (m-r)! / r! 𝔢_{m-r,r}
      coeffs[m-r+1,t+1] = ( msgn * msc^-(k+2*l+m-(m-2*r))
                          * (-1)^m * factorial(ZZ(m-r)) // factorial(ZZ(r))
                          * fcoeffs[r+1,t+1] )
    end
  else
    error()
  end

  return LRConstrSpectralPointExpansion(f.base_ring, f.poly_ring, 
           LR, -k, -l, s0+k,
           f.lpoly, f.rpoly, msc, msgn,
           coeffs)
end
